/* atomic.h */

/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef ATOMIC_H
#define ATOMIC_H


/* A barrier to prevent load and store operations from being floated around this
 * statement due to compiler optimizations.  This does not generate any CPU
 * instructions. */
#define optimization_barrier() __asm__ __volatile__("" ::: "memory")


/* Issue full memory barrier instruction, preventing any load or store
 * operations from crossing the barrier due to CPU instruction pipelining or
 * compiler optimizations. */
static inline void
memory_barrier(void)
{
	__sync_synchronize();
}


/* NEW atomic unsigned int type; intended to be opaque and only used via the
 * static inlines below.  These will be compatible with the C11 seq_cst memory
 * model when compiled in C11 mode. */
typedef struct {
	unsigned int v;
} exs_atomic_uint_t;


/* This constitutes an acquire operation.  On IA-32 and x86_64, this is
 * sufficient. */
static inline unsigned int
atomic_fetch(exs_atomic_uint_t *atom)
{
	unsigned int v;
	optimization_barrier();
	v = atom->v;
	optimization_barrier();
	return v;
}	/* atomic_fetch */

/* This constitutes a release operation; we must ensure that all prior stores
 * are made visible to other threads before this one. */
static inline void
atomic_store(exs_atomic_uint_t *atom, unsigned int new_value)
{
	memory_barrier();
	atom->v = new_value;
	optimization_barrier();
}	/* atomic_store */

static inline void
atomic_init(exs_atomic_uint_t *atom, unsigned int init_value)
{
	atom->v = init_value;
}	/* atomic_init */

static inline unsigned int
atomic_uint_add(exs_atomic_uint_t *atom, unsigned int amount)
{
	return __sync_add_and_fetch(&atom->v, amount);
}	/* atomic_uint_add */

static inline unsigned int
atomic_uint_sub(exs_atomic_uint_t *atom, unsigned int amount)
{
	return __sync_sub_and_fetch(&atom->v, amount);
}	/* atomic_uint_sub */

static inline unsigned int
atomic_uint_set_bits(exs_atomic_uint_t *atom, unsigned int mask)
{
	return __sync_fetch_and_or(&atom->v, mask);
}	/* atomic_uint_set_bits */

static inline unsigned int
atomic_uint_clear_bits(exs_atomic_uint_t *atom, unsigned int mask)
{
	return __sync_fetch_and_and(&atom->v, ~mask);
}	/* atomic_uint_clear_bits */

#endif	/* #ifndef ATOMIC_H */
