/*
 * Copyright 2007-2016 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_read_setup.h"
#include "setup/exs_send_setup.h"
#include "setup/exs_client_setup.h"
#include "setup/exs_server_setup.h"

void *test_read(void *test_void_ptr)
{
	setup_fault_catcher(p, getpid());

	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	if (p->mode > 0)
		return 0;

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_read_tests[test - 1];
		int ret =
		    exs_read(p->recv_fd, p->recv_buf, p->RECV_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* EBADF: bad fd */
	else if (test == 2) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EBADF";
		p->test_name = all_read_tests[test - 1];
		int ret = exs_read(-1, p->recv_buf, p->RECV_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EBADF);
	}

	/* ENOTCONN: socket was never connected */
	else if (test == 3) {
		p->test_value = "ENOTCONN";
		p->test_name = all_read_tests[test - 1];

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		int ret = exs_read(p->fd, p->recv_buf, p->RECV_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    ENOTCONN);
	}

	/* EINVAL: qhandle is invalid */
	else if (test == 4) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EINVAL";
		p->test_name = all_read_tests[test - 1];
		int ret = exs_read(p->recv_fd, NULL, p->RECV_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: length is not at least 1 */
	else if (test == 5) {
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		p->test_value = "EINVAL";
		p->test_name = all_read_tests[test - 1];
		int ret = exs_read(p->recv_fd, p->recv_buf, 0);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: length is greater than RDMA channel */
	else if (test == 6) {
		pid_t pid = fork();
		if (pid == 0) {
			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_bind(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_listen(p) < 0)
				exit(EXIT_FAILURE);

			int fd = open(exs_test_fifo_1, O_WRONLY);

			if (ok_exs_accept(p) < 0)
				exit(EXIT_FAILURE);

			p->test_value = "EINVAL";
			p->test_name = all_read_tests[test - 1];
			int ret =
			    exs_read(p->recv_fd, p->recv_buf, 1073741825);

			print_macro(test, p->test_value, p->test_name, ret,
				    -1, EINVAL);

			close(fd);

			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_RDONLY);

			close(fd);
			exit(EXIT_SUCCESS);
		} else {
			int fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			close(fd);

			if (ok_exs_send(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_WRONLY);

			close(fd);
		}
	}

	/* OK_EXS_READ */
	else if (test == 7) {
		pid_t pid = fork();
		if (pid == 0) {
			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_bind(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_listen(p) < 0)
				exit(EXIT_FAILURE);

			int fd = open(exs_test_fifo_1, O_WRONLY);

			if (ok_exs_accept(p) < 0)
				exit(EXIT_FAILURE);
			p->test_value = "OK_EXS_READ";
			p->test_name = all_read_tests[test - 1];
			int ret = ok_exs_read(p);

			if (ret < p->RECV_BUFSIZE) {
				print_macro(test, p->test_value,
					    p->test_name, ret, -1,
					    ENOTSUP);
			} else {
				print_macro(test, p->test_value,
					    p->test_name, ret,
					    p->RECV_BUFSIZE, 0);
			}

			close(fd);

			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_RDONLY);

			close(fd);
			exit(EXIT_SUCCESS);
		} else {
			int fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			close(fd);

			if (ok_exs_send(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_WRONLY);

			close(fd);
		}
	}

    /** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}
}

int main(int argc, char *argv[])
{
	execute(argc, argv, test_read, READ_TESTS);
}
